//**************************************************************************
//**                                                                      **
//**            Copyright 1998 by Andrew D. Pitonyak                      **
//**                                                                      **
//**   This code belongs to me, and I make no claims of functionality.    **
//**   If you desire to use it, please discuss it with me, because I am   **
//**   curious how this code is used.                                     **
//**                                                                      **
//**   Oh yes, and any bugs that you fix, please let me know...           **
//**                                                                      **
//**                                Primatech, Inc.                       **
//**   Andrew D. Pitonyak           Andrew D. Pitonyak                    **
//**   4446 Mobile Drive #105       445 Hutchinson, Suite #200            **
//**   Columbus, OH 43220           Columbus, OH 43235                    **
//**   (614)-442-8615               (614)-841-9800                        **
//**                                andyp@primatech.com                   **
//**                                                                      **
//**************************************************************************

#define INCL_DOSPROCESS       // DosPhysicalDisk()
#define INCL_DOSERRORS        // NO_ERROR, ERROR_INVALID_FUNCTION, ...
#define INCL_DOSDEVICES       // INFO_COUNT_PARTITIONABLE_DISKS

#include <os2.h>
#include <stdlib.h>
#include <string.h>
#include <exe.h>              // WORD
#include <iostream.h>
#include <iomanip.h>
#include <assert.h>
#include "ErrPrnt.hpp"

#include "DriveInf.hpp"

//**************************************************************************
//**                                                                      **
//**                                                                      **
//**************************************************************************


//**************************************************************************
//**                                                                      **
//**  Input:   APIRET* - optional                                         **
//**                                                                      **
//**  Output:  The number of partitionable drives is returned or 0        **
//**           if an error is encountered.                                **
//**           If (rc != NULL) then it is set to any error codes. If no   **
//**           error was encountered it will be set to (NO_ERROR == 0)    **
//**                                                                      **
//**************************************************************************

int numberOfPartionableDrives(APIRET* rc)
{
   ULONG    numDrives = 0;
   APIRET   error;
   error = DosPhysicalDisk(INFO_COUNT_PARTITIONABLE_DISKS, &numDrives, 2L, NULL, 0L);

   if (error != NO_ERROR)
      numDrives = 0;

   if (rc != NULL)
      *rc = error;

   return numDrives;
}


//**************************************************************************
//**                                                                      **
//**  Input:   int     - Indicates which physical hard drive to check     **
//**                     This means that 1 probably contains drive C      **
//**                     The counting is 1 based, not 0 based.            **
//**           APIRET* - optional                                         **
//**                                                                      **
//**  Output:  A file handle or 0 if an error is encountered.             **
//**           If (rc != NULL) then it is set to any error codes. If no   **
//**           error was encountered it will be set to (NO_ERROR == 0)    **
//**                                                                      **
//**  Note:    This is nothing but a wrapper for DosPhysicalDisk()        **
//**                                                                      **
//**************************************************************************

HFILE getIOControlHandle(int drive, APIRET* rc)
{
   APIRET   error;
   HFILE    handle = 0;
   int      numDrives = numberOfPartionableDrives(&error);

   if (error == NO_ERROR) {

      if (drive > 0 && drive <= numDrives) {

            char driveStr[10];
            itoa(drive, driveStr, 10);
            strcat(driveStr, ":");
            ULONG len = strlen(driveStr) + 1;

            error = DosPhysicalDisk(INFO_GETIOCTLHANDLE, &handle, 2L, driveStr, len);

      } else {

            error = ERROR_INVALID_DRIVE;

      }

   }

   if (rc != NULL)
      *rc = error;

   if (error != NO_ERROR)
      handle = 0;

   return handle;
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE   - Drive Handle to be destroyed                     **
//**                                                                      **
//**  Output:  Error code encountered, (NO_ERROR == 0) for no error       **
//**                                                                      **
//**  Note:    This is nothing but a wrapper for DosPhysicalDisk()        **
//**                                                                      **
//**************************************************************************

APIRET destroyIOControlHandle(HFILE handle)
{
   return DosPhysicalDisk(INFO_FREEIOCTLHANDLE, 0, 0, &handle, 2);
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE   - Drive Handle to be locked                        **
//**                                                                      **
//**  Output:  Error code encountered, (NO_ERROR == 0) for no error       **
//**                                                                      **
//**  Note:    This is nothing but a wrapper for DosDevIOCtl()            **
//**                                                                      **
//**************************************************************************

APIRET lockPhysicalDrive(HFILE handle)
{
   BYTE     param = 0;
   ULONG    paramLen = sizeof(param);
   ULONG    paramLenUsed = sizeof(param);
   BYTE     data = 0;
   ULONG    dataLen = sizeof(data);
   ULONG    dataLenUsed = sizeof(data);

   return DosDevIOCtl(handle, 9L, 0L, &param, paramLen, &paramLenUsed, &data, dataLen, &dataLenUsed);
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE   - Drive Handle to be unlocked                      **
//**                                                                      **
//**  Output:  Error code encountered, (NO_ERROR == 0) for no error       **
//**                                                                      **
//**  Note:    This is nothing but a wrapper for DosDevIOCtl()            **
//**                                                                      **
//**************************************************************************

APIRET unlockPhysicalDrive(HFILE handle)
{
   BYTE     param = 0;
   ULONG    paramLen = sizeof(param);
   ULONG    paramLenUsed = sizeof(param);
   BYTE     data = 0;
   ULONG    dataLen = sizeof(data);
   ULONG    dataLenUsed = sizeof(data);

   return DosDevIOCtl(handle, 9L, 1L, &param, paramLen, &paramLenUsed, &data, dataLen, &dataLenUsed);
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE   - Drive Handle                                     **
//**                                                                      **
//**  Output:  Error code encountered, (NO_ERROR == 0) for no error       **
//**           The number of cylinders, heads, and sectors per track      **
//**           are returned in cyl, heads, and sect                       **
//**                                                                      **
//**  Note:    This is nothing but a wrapper for DosDevIOCtl()            **
//**                                                                      **
//**************************************************************************

APIRET queryPhysicalDeviceParams(HFILE handle, ULONG &cyl, ULONG &heads, ULONG &sect)
{
   BYTE     param = 0;
   ULONG    paramLen = sizeof(param);
   ULONG    paramLenUsed = 1;
   WORD     data[8];
   ULONG    dataLen = sizeof(data);
   ULONG    dataLenUsed = 0;
   APIRET   error;

   error = DosDevIOCtl(handle, 9L, 0x63, &param, paramLen, &paramLenUsed, data, dataLen, &dataLenUsed);

   if (error != NO_ERROR) {

      cyl = heads = sect = 0L;

   } else {
      cyl   = data[1];
      heads = data[2];
      sect  = data[3];
      // should have disk size as (cyl * heads * sect * 512)
   }
   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   int     - Indicates which physical hard drive to check     **
//**                     This means that 1 probably contains drive C      **
//**                     The counting is 1 based, not 0 based.            **
//**                                                                      **
//**  Output:  Error code encountered, (NO_ERROR == 0) for no error       **
//**           The number of cylinders, heads, and sectors per track      **
//**           are returned in cyl, heads, and sect                       **
//**                                                                      **
//**  Note:    This is a wrapper for queryPhysicalDeviceParams() so that  **
//**           you don't need to first obtain a handle and such.          **
//**                                                                      **
//**************************************************************************

APIRET queryPhysicalDeviceParams(int drive, ULONG &cyl, ULONG &heads, ULONG &sect)
{
   APIRET   error;
   HFILE    handle = getIOControlHandle(drive, &error);
   if (error == NO_ERROR) {

      error = queryPhysicalDeviceParams(handle, cyl, heads, sect);
      //
      // Just in case an error occures in queryPhysicalDeviceParams()
      // save it so that it is not lost, this is more important than the
      // error in destroyIOControlhandle()
      //
      APIRET rc = destroyIOControlHandle(handle);
      if (error == NO_ERROR)
         error = rc;
   }
   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   const BYTE* b : Packed BYTE from the partition table       **
//**                                                                      **
//**  Output:  Extract the packed sector                                  **
//**                                                                      **
//**  Note:    This gets the sector from the packed cylinder/sector field **
//**                                                                      **
//**************************************************************************

WORD getPackedSector(const BYTE* b)
{
   return (*b) & (0x3F);
}


//**************************************************************************
//**                                                                      **
//**  Input:   const BYTE* b : Packed BYTEs from the partition table      **
//**                                                                      **
//**  Output:  Extract the packed cylinder from the partition table       **
//**                                                                      **
//**  Note:    This get the cylinder from the packed cylinder/sector field**
//**                                                                      **
//**************************************************************************

WORD getPackedCylinder(const BYTE* b)
{
   return b[1]+((b[0] & 0xC0) << 2);
}


//**************************************************************************
//**                                                                      **
//**  Input:   BYTE status : The partition status byte                    **
//**                                                                      **
//**  Output:  Ascii description of status byte                           **
//**                                                                      **
//**  Note:    Currently, only the high bit has any relevance. 1 means    **
//**           that this is the active boot partition. A 0 mean this is   **
//**           an inactive partition. The rest of the bits must be zero.  **
//**                                                                      **
//**************************************************************************

const char* partitionStatusStr(BYTE status)
{
   if (status == 0)
      return "inactive";
   else if (status == 0x80)
      return "Boot Partition";
   else
      return "unknown";
}


//**************************************************************************
//**                                                                      **
//**  Input:   BYTE aType : the byte which indentifies a partition type   **
//**                                                                      **
//**  Output:  String representing the partition table type               **
//**                                                                      **
//**  Note:                                                               **
//**                                                                      **
//**************************************************************************

const char* partitionTypeStr(BYTE aType)
{
   const char* st = "unknown";
   switch(aType) {
      case 0x0:
         st = "Empty";
         break;
      case 0x1:
         st = "FAT-12";
         break;
      case 0x2:
      case 0x3:
         st = "XENIX";
         break;
      case 0x4:
         st = "FAT-16";
         break;
      case 0x5:
         st = "Extended";
         break;
      case 0x6:
         st = "DOS >32M";
         break;
      case 0x7:
         st = "OS/2 HPFS";
         break;
      case 0x0A:
         st = "Boot Mgr";
         break;
      case 0x0B:
         st = "FAT-32";
         break;
      case 0x64:
         st = "Novell";
         break;
      case 0x75:
         st = "PCIX";
         break;
      case 0xDB:
         st = "CPM/Concurrent";
         break;
      case 0xFF:
         st = "BBT";
         break;
      default:
         break;
   }
   return st;
}


//**************************************************************************
//**                                                                      **
//**  Input:   BYTE b : a BYTE to print                                   **
//**           int includePrefix : Should the preceding 0x be printed     **
//**                                                                      **
//**  Output:  None, but the hex representation is printed                **
//**                                                                      **
//**  Note:                                                               **
//**                                                                      **
//**************************************************************************

void filledHexPrint(BYTE b, int includePrefix)
{
   if (includePrefix)
      cout << "0x";
   if (b < 0x10)
      cout << '0';
   cout << setbase(16) << (WORD) b;
}


//**************************************************************************
//**                                                                      **
//**  Input:   VOID* d : Pointer to the data to print                     **
//**           ULONG size : Number of bytes to print                      **
//**           int includePrefix : Should the preceding 0x be printed     **
//**                                                                      **
//**  Output:  None, but the data is printed                              **
//**                                                                      **
//**  Note:    Print each piece of the data in hex                        **
//**                                                                      **
//**************************************************************************

void filledHexPrint(VOID *d, ULONG size, int includePrefix)
{
   BYTE *data = (BYTE *) d;
   if (includePrefix && size > 0)
      cout << "0x";
   while (size > 0) {
      --size;
      filledHexPrint(data[size], 0);
   }
}

//**************************************************************************
//**                                                                      **
//**  Input:   None                                                       **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This simply prints the header for printing a parttion entry**
//**                                                                      **
//**************************************************************************

void prettyPrintPartitionHeader()
{
   cout << endl;
   cout << "Stat ";
   cout << "Type ";
   int len = 16;
   cout << setw(len) << " ";
   cout << "Head Cyl  Sect ";
   cout << "Head Cyl  Sect ";
   cout << "Boot Sct Num Sct";

   cout << endl;
   cout << "==== ";
   len = 20;
   for (int i=len; i>0; --i)
      cout << "=";
   cout << " ==== ==== ==== ";
   cout << "==== ==== ==== ";
   cout << "======== ======== ";
}


//**************************************************************************
//**                                                                      **
//**  Input:   const PartitionEntryStruct& pes : Reference to a partition **
//**           entry.                                                     **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This prints a partition table entry in a nice format       **
//**                                                                      **
//**************************************************************************

void prettyPrintPartitionEntry(const PartitionEntryStruct& pes)
{
   WORD tempWord;
   ULONG tempULong;
   filledHexPrint(pes.status_, 1);
   cout << " ";
   filledHexPrint(pes.type_, 1);
   int len = strlen(partitionTypeStr(pes.type_));
   cout << "=" << partitionTypeStr(pes.type_);
   if (len < 15) {
      cout << setw(15 - len) << " ";
   }
   cout << setbase(10) << setw(4) << (WORD) pes.startHead_;
   tempWord = getPackedCylinder(pes.startSecCyl_);
   cout << setbase(10) << setw(5) << tempWord;
   tempWord = getPackedSector(pes.startSecCyl_);
   cout << setbase(10) << setw(5) << tempWord;
   cout << setbase(10) << setw(5) << (WORD) pes.endHead_;
   tempWord = getPackedCylinder(pes.endSecCyl_);
   cout << setbase(10) << setw(5) << tempWord;
   tempWord = getPackedSector(pes.endSecCyl_);
   cout << setbase(10) << setw(5) << tempWord;
   cout << " " << setw(8) << setbase(10) << pes.bootSec_;
   cout << " " << setw(8) << setbase(10) << pes.numSec_;
}


//**************************************************************************
//**                                                                      **
//**  Input:   const PartitionEntryStruct* pes : Pointer to a partition   **
//**           entr                                                       **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This is simply a wrapper for the version which takes a     **
//**           reference so that a NULL pointer is not dereferenced.      **
//**                                                                      **
//**************************************************************************

void prettyPrintPartitionEntry(const PartitionEntryStruct* pes)
{
   if (pes != NULL)
      prettyPrintPartitionEntry(*pes);
}



//**************************************************************************
//**                                                                      **
//**  Input:   ULONG   - Number Of Sectors to be read                     **
//**                                                                      **
//**  Output:  None.                                                      **
//**                                                                      **
//**  Note:    Constructor                                                **
//**                                                                      **
//**************************************************************************

IOCtlParam::IOCtlParam(ULONG numSect) :
   error_(NO_ERROR), numSectors_(numSect), param_(NULL), allocatedSize_(0)
{
   initMyData();
}


//**************************************************************************
//**                                                                      **
//**  Input:   None                                                       **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    Destructor                                                 **
//**                                                                      **
//**************************************************************************

IOCtlParam::~IOCtlParam()
{
   delete[] param_;
   param_ = NULL;
   error_ = ERROR_NOT_ENOUGH_MEMORY;
}


//**************************************************************************
//**                                                                      **
//**  Input:   None                                                       **
//**                                                                      **
//**  Output:  None.                                                      **
//**                                                                      **
//**  Note:    This can be called at any time to re-initialize the data   **
//**           based on the number of sectors and such. The assumption    **
//**           is made that the data is sequential and that the sector    **
//**           size is the default size of 512.                           **
//**                                                                      **
//**************************************************************************

void IOCtlParam::initMyData()
{
   //
   // This represents the following data:
   // Byte  - Command
   // Word  - Head
   // Word  - Cylider
   // Word  - First Sector
   // Word  - Number of Sectors
   // Bytes - Track Layout
   //
   ULONG sizeRequired = sizeof(BYTE) + 4*sizeof(WORD) + 2 * sizeof(WORD)*numSectors();
   error_ = NO_ERROR;
   //
   // If the data is already allcoated and of the wrong size, the delete it
   //
   if (param_ != NULL && paramSize() != sizeRequired) {
      delete[] param_;
      allocatedSize_ = 0;
      param_ = NULL;
   }
   //
   // Allocate the actual data
   //
   if (param_ == NULL) {
      param_ = new BYTE[sizeRequired];
      if (param_ != NULL)
         allocatedSize_ = sizeRequired;
   }
   //
   // Initialize the data
   //
   if (param_ == NULL) {

      error_ = ERROR_NOT_ENOUGH_MEMORY;

   } else {

      *command() = 1;    // 0 = non-consecutive sectors, 1 = consecutive sectors
      *head() = 0;
      *cylinder() = 0;
      *firstSector() = 0;
      *numSectorsToDo() = 1;
      WORD* w = (WORD *) getOffset(sizeof(BYTE)+4*sizeof(WORD));
      for (WORD i=0; i<numSectors(); ++i) {
         w[2*i] = i+1;
         w[2*i+1] = DEFAULT_SECTOR_SIZE;
      }
   }
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE handle : File Handle such as that from               **
//**                          getIOControlHandle() which calls            **
//**                          DosPhysicalDisk(INFO_GETIOCTLHANDLE, ...    **
//**           WORD  head   : Head from which to read                     **
//**           WORD  cyl    : Cylinder from which to read                 **
//**           WORD  sect   : Sector  from which to read                  **
//**           WORD  numToRead : Number of sectors to read                **
//**           WORD  numSector : Number of sectors on a track.            **
//**           VOID* data   : The data is returned here                   **
//**           ULONG szData : How much data to read                       **
//**           IOCtlParam* param : optional,                              **
//**           WORD  sectorSize : size of each sector                     **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    Unfortunately, I have hard coded a sector size of 512      **
//**           into class IOCtlParam.                                     **
//**           There is no check that all of the data is from the same    **
//**           track which is needed for DosDecIOCtl()                    **
//**                                                                      **
//**************************************************************************

APIRET readPhysicalSectors(HFILE handle, WORD head, WORD cyl, WORD sect,
       WORD numToRead, WORD numSectors, VOID* data, ULONG szData,
       IOCtlParam* param, WORD sectorSize)
{
   APIRET error = NO_ERROR;

   int myParam = 0;
   if (param == NULL) {
      myParam = 1;
      param = new IOCtlParam(numSectors);
   }

   if (data != NULL && param != NULL && param->error() == NO_ERROR && szData >= sectorSize * numToRead){

      ULONG szParam = param->paramSize();
      ULONG szDataU = 0;
      *param->head() = head;
      *param->cylinder() = cyl;
      *param->firstSector() = sect;
      *param->numSectorsToDo() = numToRead;
      error = DosDevIOCtl(handle, 9L, 0x64, param->param(),
            param->paramSize(), &szParam, data, szData, &szDataU);

   } else {

      error = ERROR_NOT_ENOUGH_MEMORY;

   }

   if (myParam)
      delete param;

   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   int   drive  : This is 1 based so 1 = C:                   **
//**           WORD  head   : Head from which to read                     **
//**           WORD  cyl    : Cylinder from which to read                 **
//**           WORD  sect   : Sector  from which to read                  **
//**           WORD  numToRead : Number of sectors to read                **
//**           WORD  numSector : Number of sectors on a track.            **
//**           VOID* data   : The data is returned here                   **
//**           ULONG szData : How much data to read                       **
//**           IOCtlParam* param : optional,                              **
//**           WORD  sectorSize : size of each sector                     **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    Allocate a handle, call readPhysicalSectors(), and then    **
//**           free the handle                                            **
//**                                                                      **
//**************************************************************************

APIRET readPhysicalSectors(int drive, WORD head, WORD cyl, WORD sect,
       WORD numToRead, WORD numSectors, VOID* data, ULONG szData,
       IOCtlParam* param, WORD sectorSize)
{
   APIRET error = NO_ERROR;
   HFILE  handle = getIOControlHandle(drive, &error);
   if (error == NO_ERROR) {

      error = readPhysicalSectors(handle, head, cyl, sect,
              numToRead, numSectors, data, szData, param, sectorSize);

      APIRET tempError = destroyIOControlHandle(handle);
      if (error == NO_ERROR)
         error = tempError;
   }
   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE handle : File Handle                                 **
//**           WORD  head   : Head from which to read                     **
//**           WORD  cyl    : Cylinder from which to read                 **
//**           WORD  sect   : Sector  from which to read                  **
//**           WORD  sectorsPerTrack : Number of sectors on a track.      **
//**           VOID* data   : The data is returned here                   **
//**           ULONG szData : How much data to read                       **
//**           IOCtlParam* param : optional,                              **
//**           WORD  sectorSize : size of each sector                     **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    Wrapper to read a single physical sector                   **
//**                                                                      **
//**************************************************************************

APIRET readAPhysicalSector(HFILE handle, WORD head, WORD cyl, WORD sect,
       WORD sectorsPerTrack, VOID* data, IOCtlParam* param,
       WORD sectorSize)
{
   return readPhysicalSectors(handle, head, cyl, sect, 1,
               sectorsPerTrack, data, sectorSize, param, sectorSize);
}


//**************************************************************************
//**                                                                      **
//**  Input:   int   drive  : 1 = C:                                      **
//**           WORD  head   : Head from which to read                     **
//**           WORD  cyl    : Cylinder from which to read                 **
//**           WORD  sect   : Sector  from which to read                  **
//**           WORD  numSectors : Number of sectors on a track.           **
//**           VOID* data   : The data is returned here                   **
//**           ULONG szData : How much data to read                       **
//**           IOCtlParam* param : optional,                              **
//**           WORD  sectorSize : size of each sector                     **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    Wrapper to read a single physical sector                   **
//**                                                                      **
//**************************************************************************

APIRET readAPhysicalSector(int drive, WORD head, WORD cyl, WORD sect,
       WORD numSectors, VOID* data, IOCtlParam* param,
       WORD sectorSize)
{
   return readPhysicalSectors(drive, head, cyl, sect, 1,
               numSectors, data, sectorSize, param, sectorSize);
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE handle : File Handle such as that from               **
//**                          getIOControlHandle() which calls            **
//**                          DosPhysicalDisk(INFO_GETIOCTLHANDLE, ...)   **
//**           ULONG startSector : First sector to read                   **
//**           ULONG numSectors  : number of sectors to read              **
//**           VOID* data   : The data is returned here                   **
//**           IOCtlParam* param : optional,                              **
//**           WORD  sectorSize : size of each sector                     **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    This is a wrapper to call readPhysicalSectors() where the  **
//**           cylinder, head, and sector numbers are needed. These       **
//**           numbers are all zero based, even if the partition table    **
//**           enteries all seem to be one based. :-(                     **
//**           The absolute sector will be given as:                      **
//**           absSector = cyl*sects*heads + head*sects + sect            **
//**           **WARNING** There is no check that all of the data is from **
//**           the same track which is needed for DosDevIOCtl()           **
//**                                                                      **
//**************************************************************************

APIRET readPhysicalSectors(HFILE handle, ULONG startSector, ULONG numSectors,
       VOID* data, IOCtlParam* param, WORD sectorSize)
{
   APIRET   error;
   ULONG    cylinders;
   ULONG    heads;
   ULONG    sectors;

   WORD     head = 0;
   WORD     cyl  = 0;
   WORD     sect = 0;

   error = queryPhysicalDeviceParams(handle, cylinders, heads, sectors);

   if (error == NO_ERROR && numSectors > 0) {

      ULONG totSectors = cylinders * sectors * heads;
      ULONG lastSector = startSector + numSectors - 1;

      if (lastSector >= totSectors) {

         error = ERROR_INVALID_PARAMETER;

      } else {
         //
         // 0 based numbers calculated here, even if sector appears to
         // be 1 based in the partition table
         //
         ULONG uSector = startSector % sectors;
         ULONG temp = (startSector - uSector) / sectors;
         ULONG uHead = temp % heads;
         ULONG uCylinder = (temp - uHead) / heads;

         if (uHead <= 0xFFFF && uCylinder <= 0xFFFF && uSector <= 0xFFFF) {

             head = uHead;
             cyl  = uCylinder;
             sect = uSector;

         } else {

             error = ERROR_INVALID_PARAMETER;

         }
      }
      //
      // A smarter person would modify this routine so that it would work while reading data that
      // that actually spans tracks, but this satisfies my immediate needs so it will happen at a later time.
      //
      if (numSectors > 1) {

         ULONG uSector = lastSector % sectors;
         ULONG temp = (lastSector - uSector) / sectors;
         ULONG uHead = temp % heads;
         ULONG uCylinder = (temp - uHead) / heads;

         if (uSector != sect || uHead != head || uCylinder != cyl)
            error = ERROR_INVALID_PARAMETER;

      }

      if (error == NO_ERROR) {
         error = readPhysicalSectors(handle, head, cyl, sect, numSectors,
               sectors, data, sectorSize*numSectors, param, sectorSize);
      }
   }

   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   int drive : Drive to read 1: is first drive.               **
//**           PartitionSectorStruct& table : will hold the read partition**
//**                       table.                                         **
//**                                                                      **
//**  Output:  Any error number encountered                               **
//**                                                                      **
//**  Note:    This will read the partition table for the specified drive **
//**                                                                      **
//**************************************************************************

APIRET readPartitionTable(int drive, PartitionSectorStruct& table)
{
   APIRET   error;
   ULONG    cylinders;
   ULONG    heads;
   ULONG    sectors;
   HFILE    handle = getIOControlHandle(drive, &error);
   WORD     head = 0;
   WORD     cylinder = 0;
   WORD     sector = 0;

   if (error == NO_ERROR) {
      error = queryPhysicalDeviceParams(handle, cylinders, heads, sectors);
      if (error == NO_ERROR) {
         error = readAPhysicalSector(handle, head, cylinder, sector,
               sectors, (BYTE *) &table, NULL, DEFAULT_SECTOR_SIZE);
      }
      destroyIOControlHandle(handle);
   }

   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   void* d : This is the data to dump to the screen           **
//**           ULONG thisSize : How big is the data                       **
//**           int   zeroPad  : Print small numbers with a leading zero   **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This will print a HEX dump to the screen                   **
//**                                                                      **
//**************************************************************************

void dumpSector(VOID* d, ULONG thisSize, int zeroPad)
{
   BYTE* data = (BYTE *) d;
   if (data != NULL && thisSize > 0) {
      ULONG done = 0;
      int row = 0;
      while (done < thisSize) {
         cout << endl << setbase(16) << setw(4) << row << " - ";
         for (int i=0; i<16 && done < thisSize; ++i) {
            if (zeroPad && data[done] < 0x10) {
                cout << "0" << setw(1) << (int) data[done] << " ";
            } else {
                cout << setw(2) << (int) data[done] << " ";
            }
            ++done;
         }
         row += 16;
      }
   }
}


//**************************************************************************
//**                                                                      **
//**  Input:   int drive : 1 based drive to read                          **
//**           ULONG startSector : First sector to read                   **
//**           ULONG numSectors  : number of sectors to read              **
//**                                                                      **
//**  Output:  Any returned error codes                                   **
//**                                                                      **
//**  Note:    read the data and simply dump the sectors                  **
//**                                                                      **
//**************************************************************************

APIRET readAndDumpSectors (int drive, ULONG startSector, ULONG numSectors)
{
   ULONG dataSize = 1024;
   BYTE* data = new BYTE[dataSize];

   APIRET error = NO_ERROR;
   ULONG lastSector = startSector + numSectors;
   ULONG currentSector = startSector;

   HFILE  handle = getIOControlHandle(drive, &error);

   if (error == NO_ERROR) {
       while (currentSector < lastSector && error == NO_ERROR) {
           error = readPhysicalSectors(handle, currentSector, 1, data);
           if (error == NO_ERROR) {

               cout << endl << endl << "Drive " << drive << ": sector " << currentSector << endl;
               dumpSector(data, DEFAULT_SECTOR_SIZE, 1);

           } else {

               printErrorMessage(error, "Error reading physical sectors");

           }
           ++currentSector;
       }
       //
       // Note that this is destroyed only if it was successfully  created
       //
       APIRET tempError = destroyIOControlHandle(handle);
       if (error == NO_ERROR)
           error = tempError;

       if (tempError != NO_ERROR)
           printErrorMessage(error, "Error from destroyIOControlHandle()");

   } else {

       printErrorMessage(error, "Error from getIOControlHandle()");

   }
   delete[] data;
   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   HFILE handle : Already opened handle to a drive            **
//**           ULONG absSector : Sector from which to read the table      **
//**           ULONG offset : Current offset to add to sectors to find    **
//**                          the offset of extended partitions. see note **
//**                                                                      **
//**  Output:  Any errors encountered                                     **
//**                                                                      **
//**  Note:    This will print the partition table information. It will   **
//**           then recursively print the partition information for each  **
//**           extended partition. The problem is that precisely how the  **
//**           information is not all that consistent (in my opinion).    **
//**           Heres the scoop. For the primary partition table the       **
//**           offset will be zero so the start sector of the             **
//**           partitions will be relative to this (or absolute).         **
//**           When I enter an extended partition, the offsets are        **
//**           relative to the start of this new partition. So far, it is **
//**           easy. The extended partition will frequently store logical **
//**           disks which it contains as though they are stored inside   **
//**           of an extended partition inside of the current extended    **
//**           partition. The problem is that the values inside of these  **
//**           fake extended partition tables are relative to the start   **
//**           of the initial extended partition as opposed to the        **
//**           extended partition which is being read. This caused me     **
//**           grief and is the reason for the strange use of "offset"    **
//**                                                                      **
//**************************************************************************

APIRET printPartitionInformation(HFILE handle, ULONG absSector, ULONG offset)
{
   APIRET error = NO_ERROR;
   PartitionSectorStruct *table = new PartitionSectorStruct;

   if (table != NULL)
      error = readPhysicalSectors(handle, absSector, 1, table);
   else
      error = ERROR_NOT_ENOUGH_MEMORY;

   if (error == NO_ERROR) {
      //
      // Verify that the proper id is present. It is possible that
      // you will want to print the information anyway. For when you
      // are looking at dammaged partition tables for example.
      //
      if (table->iDCode_ != 0xAA55) {
         error = ERROR_GEN_FAILURE;
         cout << endl << "Error, bad iDCode, found ";
         filledHexPrint(&table->iDCode_, sizeof(table->iDCode_), 1);
         cout << endl;
      }
      //
      // Print each entry
      //
      cout << endl << "Partition at offset " << absSector;
      prettyPrintPartitionHeader();
      for (int i=0; i<4; i++) {
         cout << endl;
         prettyPrintPartitionEntry(table->entries_[i]);
      }
      cout << endl;
   }
   //
   // Recusively print each extended partition type
   //
   int i = 0;
   while (error == NO_ERROR && i < 4) {
      //
      // If the type is 0x05 (extended) ...
      //
      if (table->entries_[i].type_ == 0x05) {
         ULONG sect = table->entries_[i].bootSec_;
         //
         // Do we take the offset from this partition table, or
         // from the already established offset.
         //
         ULONG newOffset = offset == 0 ? sect : offset;
         error = printPartitionInformation(handle, sect + offset, newOffset);
      }
      ++i;
   }

   delete table;
   table = NULL;

   return error;
}


//**************************************************************************
//**                                                                      **
//**  Input:   long drive : drive number                                  **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This prints number of heads, cyliers, and sectors          **
//**                                                                      **
//**************************************************************************

void test_queryPhysicalDeviceParams(long drive)
{
   APIRET error;
   int first = 1;
   int last  = numberOfPartionableDrives(&error);
   if (first <= drive && drive <= last) {
      first = last = drive;
   }

   for (int i = first; i <= last; ++i) {
      cout << endl << "*********************************************************";
      cout << endl << "queryPhysicalDeviceParams() for physical drive " << i << endl;
      HFILE    handle = getIOControlHandle(i, &error);
      if (error == NO_ERROR) {
         ULONG cyl = 0;
         ULONG heads = 0;
         ULONG sect = 0;
         error = queryPhysicalDeviceParams(1, cyl, heads, sect);
         cout << endl << "cyl = " << cyl << " heads = " << heads << " sect = " << sect << endl;
         if (error != NO_ERROR) {
            printErrorMessage(error, " from queryPhysicalDeviceParams()");
         }
         error = destroyIOControlHandle(handle);
         if (error != NO_ERROR)
            printErrorMessage(error, " from destroyIOControlHandle()");
      } else {
         printErrorMessage(error, " from getIOControlHandle()");
      }
   }
}


//**************************************************************************
//**                                                                      **
//**  Input:   long drive                                                 **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    Given a drive, this prints information about all of the    **
//**           partitions on this drive.                                  **
//**                                                                      **
//**************************************************************************

void test_printPartitionInfo(long drive)
{
   APIRET error;
   int first = 1;
   int last  = numberOfPartionableDrives(&error);
   if (first <= drive && drive <= last) {
      first = last = drive;
   }

   for (int i = first; i <= last; ++i) {
      cout << endl << "Partition info for physical drive " << i << endl;
      cout << "*********************************************************" << endl;
      HFILE    handle = getIOControlHandle(i, &error);
      if (error == NO_ERROR) {
         error = printPartitionInformation(handle, 0L, 0L);
         if (error != NO_ERROR)
            printErrorMessage(error, " from printPartitionInformation()");
         error = destroyIOControlHandle(handle);
         if (error != NO_ERROR)
            printErrorMessage(error, " from destroyIOControlHandle()");
      } else {
         printErrorMessage(error, " from getIOControlHandle()");
      }
   }
}


//**************************************************************************
//**                                                                      **
//**  Input:   None                                                       **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    This prints the number of partitionable drives             **
//**           Simply a wrapper function                                  **
//**                                                                      **
//**************************************************************************

void test_numberOfPartionableDrives()
{
   APIRET error;
   int numDrives = numberOfPartionableDrives(&error);
   if (error == NO_ERROR)
      cout << endl << "Number of partitionable drives = " << numDrives << endl;
   else
      printErrorMessage(error, " from numberOfPartionableDrives()");
}


//**************************************************************************
//**                                                                      **
//**  Input:   None                                                       **
//**                                                                      **
//**  Output:  None                                                       **
//**                                                                      **
//**  Note:    Print a bit of help                                        **
//**                                                                      **
//**************************************************************************

void instruct()
{
   cout << endl << "Drive Information Usage" << endl;
   cout << endl << "Drives can be referenced as either a physical or a logical drive.";
   cout << endl << "Physical drives are referenced as 1:, 2:... and logical drives as.";
   cout << endl << "C:, D:...";
   cout << endl << "Usage:";
   cout << endl << "num physical - Prints number of physical drives";
   cout << endl << "partition print {#} - If # is not entered then all are printed";
   cout << endl << "params physical {#} - If # is not entered then all are printed";
   cout << endl << "dump {#physical drive (a based)} {#first sector (0 based)} {#numsectors}";
   cout << endl;
}


//**************************************************************************
//**                                                                      **
//**  Input:   arguments                                                  **
//**                                                                      **
//**  Output:  0                                                          **
//**                                                                      **
//**  Note:    An interface to my test code                               **
//**                                                                      **
//**************************************************************************

int main(int argc, char *argv[])
{

   cout << endl << "Program by Andrew D. Pitonyak, andyp@primatech.com";
   cout << endl;
   int printHelp = argc < 2 || argv[1] == NULL;
   int ok = !printHelp;
   int finished = !ok;
   long num;

   if (!finished && ok && strcmpi(argv[1], "dump") == 0) {
      int drive = argc > 1 ? atol(argv[2]) : 1;
      ULONG firstSector = argc > 2 ? atol(argv[3]) : 0;
      ULONG numSectors = argc > 3 ? atol(argv[4]) : 1;
      readAndDumpSectors (drive, firstSector, numSectors);
      finished = 1;
   }

   if (!finished && ok && strcmpi(argv[1], "num") == 0) {
      if (argc > 2 && strcmpi(argv[2], "physical") == 0) {
         test_numberOfPartionableDrives();
         finished = 1;
      }
   }

   if (!finished && ok && strcmpi(argv[1], "partition") == 0) {
      if (argc > 2 && strcmpi(argv[2], "print") == 0) {
         num = argc > 3 ? atol(argv[3]) : -1;
         test_printPartitionInfo(num);
         finished = 1;
      }
   }

   if (!finished && ok && strcmpi(argv[1], "params") == 0) {
      if (argc > 2 && strcmpi(argv[2], "physical") == 0) {
         num = argc > 3 ? atol(argv[3]) : -1;
         test_queryPhysicalDeviceParams(num);
         finished = 1;
      }
   }


   if (printHelp || !finished)
      instruct();

   return 0;
}
