//**************************************************************************
//**                                                                      **
//**            Copyright 1998 by Andrew D. Pitonyak                      **
//**                                                                      **
//**   This code belongs to me, and I make no claims of functionality.    **
//**   If you desire to use it, please discuss it with me, because I am   **
//**   curious how this code is used.                                     **
//**                                                                      **
//**   Oh yes, and any bugs that you fix, please let me know...           **
//**                                                                      **
//**                                Primatech, Inc.                       **
//**   Andrew D. Pitonyak           Andrew D. Pitonyak                    **
//**   4446 Mobile Drive #105       445 Hutchinson, Suite #200            **
//**   Columbus, OH 43220           Columbus, OH 43235                    **
//**   (614)-442-8615               (614)-841-9800                        **
//**                                andyp@primatech.com                   **
//**                                                                      **
//**************************************************************************
#if !defined(DriveInf_hpp__)
#define DriveInf_hpp__

#include "DrvStrct.hpp"

//**************************************************************************
//**                                                                      **
//**                                                                      **
//**************************************************************************

//**************************************************************************
//**                                                                      **
//**
//**                                                                      **
//**************************************************************************

class IOCtlParam {
public:
   IOCtlParam(ULONG numSect);
   virtual ~IOCtlParam();

   APIRET error() {return error_;}
   BYTE*  param() const {return param_;}
   ULONG  paramSize() const {return allocatedSize_;}
   ULONG& numSectors() {return numSectors_;}

   BYTE* command() const {return param_;}
   WORD* head() const {return (WORD *) getOffset(sizeof(BYTE));}
   WORD* cylinder() const {return (WORD *) getOffset(sizeof(BYTE)+sizeof(WORD));}
   WORD* firstSector() const {return (WORD *) getOffset(sizeof(BYTE)+2*sizeof(WORD));}
   WORD* numSectorsToDo() const {return (WORD *) getOffset(sizeof(BYTE)+3*sizeof(WORD));}

protected:
   BYTE* getOffset(int o) const {return param_ != NULL ? &param_[o] : NULL;}
   virtual void fillParam() {initMyData();}
   void initMyData();

protected:
   ULONG    allocatedSize_;
   APIRET   error_;
   ULONG    numSectors_;
   BYTE    *param_;
};





int numberOfPartionableDrives(APIRET* rc = 0);
HFILE getIOControlHandle(int drive, APIRET* rc = 0);
APIRET destroyIOControlHandle(HFILE handle);
APIRET lockPhysicalDrive(HFILE handle);
APIRET unlockPhysicalDrive(HFILE handle);
APIRET queryPhysicalDeviceParams(HFILE handle, ULONG &cyl, ULONG &heads, ULONG &sect);
APIRET queryPhysicalDeviceParams(int drive, ULONG &cyl, ULONG &heads, ULONG &sect);
WORD getPackedSector(const BYTE* b);
WORD getPackedCylinder(const BYTE* b);
const char* partitionStatusStr(BYTE status);
const char* partitionTypeStr(BYTE aType);
void filledHexPrint(BYTE b, int includePrefix = 0);
void filledHexPrint(VOID *d, ULONG size, int includePrefix = 0);
void prettyPrintPartitionHeader();
void prettyPrintPartitionEntry(const PartitionEntryStruct& pes);
void prettyPrintPartitionEntry(const PartitionEntryStruct* pes);
APIRET readPhysicalSectors(HFILE handle, WORD head, WORD cyl, WORD sect,
       WORD numToRead, WORD numSectors, VOID* data, ULONG szData,
       IOCtlParam* param=NULL, WORD sectorSize = DEFAULT_SECTOR_SIZE);
APIRET readPhysicalSectors(int drive, WORD head, WORD cyl, WORD sect,
       WORD numToRead, WORD numSectors, VOID* data, ULONG szData,
       IOCtlParam* param=NULL, WORD sectorSize = DEFAULT_SECTOR_SIZE);
APIRET readAPhysicalSector(HFILE handle, WORD head, WORD cyl, WORD sect,
       WORD sectorsPerTrack, VOID* data, IOCtlParam* param=NULL,
       WORD sectorSize = DEFAULT_SECTOR_SIZE);
APIRET readAPhysicalSector(int drive, WORD head, WORD cyl, WORD sect,
       WORD numSectors, VOID* data, IOCtlParam* param=NULL,
       WORD sectorSize = DEFAULT_SECTOR_SIZE);
APIRET readPhysicalSectors(HFILE handle, ULONG startSector, ULONG numSectors,
       VOID* data, IOCtlParam* param=NULL, WORD sectorSize = DEFAULT_SECTOR_SIZE);
APIRET readPartitionTable(int drive, PartitionSectorStruct& table);
void dumpSector(VOID* d, ULONG thisSize = DEFAULT_SECTOR_SIZE, int zeroPad=0);
APIRET printPartitionInformation(HFILE handle, ULONG absSector);




#endif

